/* Preference.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;

import java.text.DecimalFormat;
import java.util.InputMismatchException;
import java.util.Scanner;

import org.apache.bsf.BSFException;
import org.apache.bsf.BSFManager;

import weka.estimators.ConditionalEstimator;
import weka.estimators.Estimator;

public class Preference {
	
	private enum FunctionFormat {NONE, SINGLEVALUE, STEPFUNCTION, EQUATION, ESTIMATOR};
	

	/** The name of the attribute with the preference */
	private String m_attName = null;

	/** Setter for the attribute name */
	public void setAttributeName(String attName) { m_attName = attName; }

	/** Getter for the attribute name */
	public String getAttributeName() { return m_attName; }

	
	
	/** The weight of this preference */
	private double m_weight = 1.0;

	/** Setter for the weight of this preference.
	 *  @param weight The weight of this preference.
	 */
	public void setWeight(double weight) {
		if (weight < 0)
			throw new IllegalArgumentException("Weight must be nonnegative.");
		m_weight = weight;
	}

	/** Getter for the weight of this preference.
	 *  @return The weight of this preference.
	 */
	public double weight() { return m_weight; }
	
	
	
	private FunctionFormat m_diversityFnFormat = FunctionFormat.NONE;
	
	/** The preferred diversity, NaN if not specified */
	private double m_diversity_targetValue = Double.NaN;

	/** Setter for the preferred diversity as a single value.
	 * @param value The preferred diversity.
	 */
	public void setDiversityFn(double value) {
		if (!valueInRange(value, 0, 1))
			throw new IllegalArgumentException("Diversity must be between 0 and 1.");
		m_diversityFnFormat = FunctionFormat.SINGLEVALUE;
		m_diversity_targetValue = value;
	}
	
	
	private KernelDensityEstimator m_diversity_estimator = null;
	
	/** Setter for the diversity function based on an estimator.
	 * @param estimator The estimator for the diversity
	 */
	public void setDiversityFn(KernelDensityEstimator estimator) {
		if (estimator == null)
			throw new IllegalArgumentException("Diversity function cannot be null.");
		m_diversityFnFormat = FunctionFormat.ESTIMATOR;
		m_diversity_estimator = estimator;
	}

	/** Getter for the relative diversity match.
	 * @return The match between the target and given diversity values.  1 indicates a perfect match, 0 a poor match.
	 */
	public double diversityFn(double diversityValue) {
		if (!hasDiversity())
			return 1;
			//throw new IllegalStateException("Diversity preference not specified");
		
		if (m_diversityFnFormat == FunctionFormat.SINGLEVALUE) {
			// 1 minus the squared error between the two values
			return 1.0 - Math.pow(Math.abs(diversityValue - m_diversity_targetValue),2);
		}
		else if (m_diversityFnFormat == FunctionFormat.ESTIMATOR) {
			return m_diversity_estimator.getProbability(diversityValue);
		}
		
		throw new IllegalStateException("Diversity function not specified.");
	}
	
	/** Getter for the peak of the diversity function
	 */
	public double diversityFnPeak() {
		if (!hasDiversity())
			return Double.NaN;
		
		switch (m_diversityFnFormat) {
			case SINGLEVALUE:	return m_diversity_targetValue;
			case ESTIMATOR:		return m_diversity_estimator.getPeak();
			default:			throw new IllegalStateException("Diversity function not specified.");
		}
	}

	/** Check whether there is a diversity preference.
	 * @return <code>true</code> if there is a diversity preference,
	 *         <code>false</code> otherwise.
	 */
	public boolean hasDiversity() { return (m_diversityFnFormat != FunctionFormat.NONE); }

	/** Unsets the diversity preference. */
	public void unsetDiversity() {
		m_diversityFnFormat = FunctionFormat.NONE;
		m_diversity_targetValue = Double.NaN;
		m_diversity_estimator = null;
	}


	
	

	private FunctionFormat m_qualityFnFormat = FunctionFormat.NONE;
	
	
	/** The max value, NaN if not specified */
	private double m_qualityFn_maxValue = Double.NaN;

	/** The min value, NaN if not specified */
	private double m_qualityFn_minValue = Double.NaN;
	
	/** Sets the value function to be a step function.
	 * @param minValue The min value of the range that has a value of 1.0;
	 * @param maxValue The max value of the range that has a value of 1.0;
	 */
	public void setQualityFn(double minValue, double maxValue) {
		if (Double.isNaN(minValue) ||
			Double.isNaN(maxValue) ||
			minValue > maxValue)
			throw new IllegalArgumentException("Improper arguments: minValue="+minValue+", maxValue="+maxValue+".");
				
		m_qualityFnFormat = FunctionFormat.STEPFUNCTION;
		m_qualityFn_minValue = minValue;
		m_qualityFn_maxValue = maxValue;
	}

	
	
	/** The value function for this attribute, null if unspecified */
	private String m_qualityFn_EquationCodeString = null;
	
	protected double[] m_distributionProbs = null;
	protected double m_minXCoord = 0;
	protected double m_maxXCoord = 100;
	protected double m_offset = 0.01;

	/** Setter for the value function.
	 * @param codeString
	 *            The value function as a javascript string. The value function
	 *            should compute a value based on the variable named
	 *            <code>x</code>. During runtime, the value specified to
	 *            evaluate the value function will be substituted for
	 *            <code>x</code>. Example codeStrings: "x + 1", "Math.pow(x,
	 *            2) / 4".
	 */
	public void setQualityFn(String codeString) {
		
		m_qualityFnFormat = FunctionFormat.EQUATION;
		
		// try the value function on an arbitrary input to see if it parses validly
		try {
			BSFManager scriptManager = new BSFManager();
			scriptManager.declareBean("x", new Double(1.0), Double.class);
			Object result = scriptManager.eval("javascript", "ValueFn", 0, 0,
					codeString);
			(new Double(result.toString())).doubleValue();
		} catch (Exception e) {
			System.err.println(e.getMessage());
			throw new IllegalArgumentException(
					"Specified codeString is invalid javascript: \""
							+ codeString + "\".");
		}
	
		// set up the lookup table for the values
		double x = 0;
		try {
			// generate the probabilities over this range
			int numBins = (int) Math.ceil((m_maxXCoord - m_minXCoord) / m_offset) + 1;
			m_distributionProbs = new double[numBins];
			for (int i = 0; i < numBins; i++) {
				x = m_minXCoord + i*m_offset;
				BSFManager scriptManager = new BSFManager();
				scriptManager.declareBean("x", new Double(x), Double.class);
				Object result = scriptManager.eval("javascript", "ValueFn", 0, 0, codeString);
				m_distributionProbs[i] = (new Double(result.toString())).doubleValue();
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			throw new RuntimeException("An error occurred when evaluating the value function "+codeString+" with x="+x+".");
		}
		
		// store the value function code
		m_qualityFn_EquationCodeString = codeString;
	}
	
	
	/** The value function estimator, null if unspecified */
	private Estimator m_qualityFn_Estimator = null;
	
	/** Setter for the value function
	 *  @param estimator The estimator for the value function.
	 */
	public void setQualityFn(Estimator estimator) { 
		
		if (estimator == null)
			throw new IllegalArgumentException("Estimator cannot be null.");
		
		m_qualityFnFormat = FunctionFormat.ESTIMATOR;
		m_qualityFn_Estimator = estimator;
	}

	/** Unsets the value function. */
	public void unsetQualityFn() {
		m_qualityFnFormat = FunctionFormat.NONE;
		m_qualityFn_EquationCodeString = null;
		m_qualityFn_Estimator = null;
		m_distributionProbs = null;
		m_qualityFn_minValue = Double.NaN;
		m_qualityFn_maxValue = Double.NaN;
	}

	/** Check whether there is a value function defined.
	 * @return <code>true</code> if there is a value function defined,
	 *         <code>false</code> otherwise.
	 */
	public boolean hasQualityFn() { return (m_qualityFnFormat != FunctionFormat.NONE); }

	/** Evaluates the quality function with the specified value.
	 * @param value The value to evaluate.
	 * @return The result of the value function with the specified value.
	 * @see setValueFn
	 */
	public double qualityFn(double value) {

		if (!hasQualityFn()) {
			throw new IllegalStateException("Quality function not specified.");
		}
		
		
		// step function
		if (m_qualityFnFormat == FunctionFormat.STEPFUNCTION) {
			if (value >= m_qualityFn_minValue && value <= m_qualityFn_maxValue)
				return 1;
			else
				return 0;
		}
		
		
		// estimator
		if (m_qualityFnFormat == FunctionFormat.ESTIMATOR) {
			return m_qualityFn_Estimator.getProbability(value);
		}
		
		
		// equation
		if (m_qualityFnFormat == FunctionFormat.EQUATION) { 
			double fnValue = Double.NaN;

			// regenerate the distribution if x is outside the specified bounds
			if (value > m_maxXCoord || value < m_minXCoord) {
				// resize the appropriate bound
				if (value > m_maxXCoord) { m_maxXCoord = value + 2*(value - m_maxXCoord); }
				else { m_minXCoord = value - 2*(m_minXCoord - value); }
				// regenerate the value function
				setQualityFn(m_qualityFn_EquationCodeString);
			}
			// figure out the index into the distribution
			double roundedValue = Math.rint(value / m_offset) * m_offset;
			int idx = (int) Math.rint((roundedValue - m_minXCoord) / m_offset);
			if (roundedValue == value) {
				fnValue = m_distributionProbs[idx];
			} else {
				idx = (roundedValue > value) ? idx-1 : idx;
				// interpolate the value
				// y = (([idx+1]-[idx])/offset)*(x-(min+offset*idx))+[idx]
				try {
					fnValue = (m_distributionProbs[idx+1] - m_distributionProbs[idx])/m_offset *
						(value - (m_minXCoord+m_offset*idx)) + m_distributionProbs[idx];
				} catch (ArrayIndexOutOfBoundsException e) {
					System.err.println("idx = "+idx+", len = "+m_distributionProbs.length+", value = "+value+", roundedValue = "+roundedValue);
					e.printStackTrace();
					throw e;
				}
			}
		
			if (!valueInRange(fnValue,0,1)) {
				System.err.println("Value function "+m_qualityFn_EquationCodeString+" can return values outside of [0,1].  For example, "+m_qualityFn_EquationCodeString+" = "+fnValue+" for x = "+value);
				System.err.println("Clipping value for x = "+value+" to [0,1]");
				if (fnValue > 1.0) fnValue = 1.0;
				else if (fnValue < 0) fnValue = 0;
			}
			return fnValue;
		}
		
		throw new IllegalStateException("Quality function not specified.");
	}


	/** Creates a preference from a string representation of that preference
	 *  @param s The string representation of the preference.
	 */
	public Preference(String s) {
		try {
			Scanner lineScanner = new Scanner(s);
			// initialize the preference's attribute
			setAttributeName(lineScanner.next());
			// read in the preference tokens
			double valueFn_min = Double.NaN;
			double valueFn_max = Double.NaN;
			while (lineScanner.hasNext()) {
				String token = lineScanner.next();
				if (token.equalsIgnoreCase(":diversity")) {
					setDiversityFn(lineScanner.nextDouble());
				} else if (token.equalsIgnoreCase(":min")) {
					valueFn_min = lineScanner.nextDouble();
				} else if (token.equalsIgnoreCase(":max")) {
					valueFn_max = lineScanner.nextDouble();
				} else if (token.equalsIgnoreCase(":weight")) {
					setWeight(lineScanner.nextDouble());
				} else if (token.equalsIgnoreCase(":valuefn")) {
					lineScanner.findInLine("\".*\"");
					String commandString = lineScanner.match().group(0);
					// strip the leading and trailing quotation marks
					commandString = commandString.substring(1, commandString
							.length() - 1);
					setQualityFn(commandString);
				} else {
					throw new IllegalArgumentException("Invalid preference string: \"" + s + "\".");
				}
			}
			
			if (!Double.isNaN(valueFn_min) || !Double.isNaN(valueFn_max)) {
				if (Double.isNaN(valueFn_min))
					valueFn_min = Double.NEGATIVE_INFINITY;
				if (Double.isNaN(valueFn_max))
					valueFn_max = Double.POSITIVE_INFINITY;
				setQualityFn(valueFn_min,valueFn_max);
			}
			
		} catch (InputMismatchException e) {
			throw new IllegalArgumentException("Invalid preference string: \""
					+ s + "\".");
		} catch (Exception e) {
			System.err.println(e.getMessage());
			throw new IllegalArgumentException("Invalid preference string: \""
					+ s + "\".");
		}
	}

	public String toString() {
		DecimalFormat df = new DecimalFormat("0.0000");
		String s = getAttributeName();
		s += " :weight " + df.format(m_weight);
		if (hasDiversity()) {
			if (m_diversityFnFormat == FunctionFormat.SINGLEVALUE)
				s += " :diversity " + df.format(m_diversity_targetValue);
			else if (m_diversityFnFormat == FunctionFormat.ESTIMATOR)
				s += " :diversityfnEst \"" + m_diversity_estimator + "\"";
		}
		if (hasQualityFn()) {
			if (m_qualityFnFormat == FunctionFormat.STEPFUNCTION) {
				s += " :min " + df.format(m_qualityFn_minValue);
				s += " :max " + df.format(m_qualityFn_maxValue);
			}
			else if (m_qualityFnFormat == FunctionFormat.EQUATION)
				s += " :valuefn \"" + m_qualityFn_EquationCodeString + "\"";
			else if (m_qualityFnFormat == FunctionFormat.ESTIMATOR)
				s += " :valuefnEst \"" + m_qualityFn_Estimator + "\"";
		}
		return s;
	}

	public Preference() {
	}
	
	public Preference(Preference p) {
		this();
		this.setWeight(p.weight());
		this.setAttributeName(p.getAttributeName());
		m_diversityFnFormat = p.m_diversityFnFormat;
		if (p.hasDiversity()) {
			if (p.m_diversityFnFormat == FunctionFormat.SINGLEVALUE)
				this.setDiversityFn(p.m_diversity_targetValue);
			else if (p.m_qualityFnFormat == FunctionFormat.ESTIMATOR)
				this.setDiversityFn(p.m_diversity_estimator);
		}
		m_qualityFnFormat = p.m_qualityFnFormat;
		if (p.hasQualityFn()) {
			if (p.m_qualityFnFormat == FunctionFormat.STEPFUNCTION)
				this.setQualityFn(p.m_qualityFn_minValue, p.m_qualityFn_maxValue);
			else if (p.m_qualityFnFormat == FunctionFormat.EQUATION)
				this.setQualityFn(p.m_qualityFn_EquationCodeString);
			else if (p.m_qualityFnFormat == FunctionFormat.ESTIMATOR)
				this.setQualityFn(p.m_qualityFn_Estimator);
		}
	}

	/** Checks that a value is within the specified range.
	 * @param value The value to check.
	 * @param minValue The minimum value allowed.
	 * @param maxValue The maximum value allowed.
	 * @return <code>true</code> if minValue <= value <= maxValue,
	 *         <code>false</code> otherwise.
	 */
	private static boolean valueInRange(double value, double minValue, double maxValue) {
		double EPSILON = 1.0E-10;
		return (value >= (minValue - EPSILON) && value <= (maxValue + EPSILON));
	}

}
