/* Preferences.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;

public class Preferences implements Iterable{

	HashMap<String, Preference> m_prefMap = null;

	boolean m_weightSumComputed = false;

	private double m_weightSum = Double.NaN;
	
	private double m_alpha = Double.NaN;

	String m_name = null;

	public Preferences(String name) {
		m_prefMap = new HashMap<String, Preference>();
		m_name = name;
	}

	/**
	 * Reads preferences from a file, and loads them into the preferences
	 * structure Example file format:
	 * <code>:alpha 0.5</code>
	 * <code>size :diversity 0.5 :depth 0.5</code>
	 * <code>length :min 1 :max 5</code>
	 * <code>cost :valuefn "1 - x/1000000"</code>
	 * 
	 * @param preferencesFile
	 *            The file containing the preferences.
	 */
	public Preferences(String name, File preferencesFile) {
		this(name);
		System.out.println("Preferences " + name + ":");
		try {
			BufferedReader fileReader = new BufferedReader(new FileReader(
					preferencesFile));

			// read line by line throught the file
			String line;
			while ((line = fileReader.readLine()) != null) {
				try {
					// skip comments
					if (line.startsWith("#"))
						continue;
					
					// read in alpha
					if (line.startsWith(":alpha")) {
						String[] lineSplit = line.split(" ");
						try {
							setAlpha(Double.parseDouble(lineSplit[1]));
						} catch (Exception e) {
							throw new IllegalArgumentException("Unspecified value for alpha");
						}
						continue;
					}

					Preference p = new Preference(line);
					add(p);
					System.out.println("Added preference for " + p);
				} catch (IllegalArgumentException e) {
					System.err.println("Invalid line: \"" + line
							+ "\".  Skipping.");
				}
			}
		} catch (FileNotFoundException e) {
			System.err
					.println("getPreferencesFromFile() did not complete successfully.");
			System.err.println(e.getMessage());
		} catch (IOException e) {
			System.err
					.println("getPreferencesFromFile() did not complete successfully.");
			System.err.println(e.getMessage());
		}
		
		// check for specified alpha value
		if (Double.isNaN(alpha())) {
			throw new IllegalStateException("No alpha value specified.");
		}
	}

	public void add(Preference p) {
		m_prefMap.put(p.getAttributeName(), p);
		m_weightSumComputed = false;
	}

	public Preference get(String attributeName) {
		return m_prefMap.get(attributeName);
	}

	public boolean hasPreference(String attributeName) {
		return m_prefMap.containsKey(attributeName);
	}

	public double weightSum() {
		if (!m_weightSumComputed) {
			// compute the weight sum
			double weightSum = 0;
			for (Preference p : m_prefMap.values()) {
				weightSum += p.weight();
			}
			m_weightSum = weightSum;
			m_weightSumComputed = true;
		}
		// return the weight sum
		return m_weightSum;
	}

	public void remove(String attributeName) {
		m_prefMap.remove(attributeName);
		m_weightSumComputed = false;
	}

	public Iterator<Preference> iterator() {
		return m_prefMap.values().iterator();
	}

	public void clear() {
		m_prefMap.clear();
		m_weightSumComputed = true;
		m_weightSum = 0;
	}

	public int size() {
		return m_prefMap.size();
	}

	public String name() {
		return m_name;
	}
	
	public void setAlpha(double alpha) {
		m_alpha = alpha;
	}
	
	public double alpha() {
		return m_alpha;
	}

	public String toString() {
		String s = "Preferences: num=" + size() + ".\n";
		s += ":alpha "+alpha()+"\n";
		for (Preference p : m_prefMap.values()) {
			s += p.toString() + "\n";
		}
		return s;
	}

	public static void main(String[] args) {
		Preferences prefs = new Preferences("blocks-mosaicvfn", new File(
				"blocks-mosaicvfn.pref"));
		System.out.println(prefs);
	}
}
