/* SingleStepEstimator.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


package ddpref;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;

import weka.estimators.Estimator;

/** 
 * Simple single step-function estimator. Estimates a function with a
 * single step that ranges between 0 and 1.0.
 *
 * @author Eric Eaton
 * @version $Revision: 1.2 $
 */
public class SingleStepEstimator implements Estimator {
	
	/** Tracks the min value added */
	protected double m_minValue = Double.POSITIVE_INFINITY;
	
	/** Tracks the max value added */
	protected double m_maxValue = Double.NEGATIVE_INFINITY;
	
	protected ArrayList<Double> m_data = new ArrayList<Double>();

	/** Constructor */
	public SingleStepEstimator() {
		super();
	}

	/** Adds a value to the distribution.  The weight is ignored.
	 * @param data The data value to add.
	 * @param weight The weight of the data value.  Ignored.
	 */
	public void addValue(double data, double weight) {
		// update the min and max values
		m_minValue = (data < m_minValue) ? data : m_minValue;
		m_maxValue = (data > m_maxValue) ? data : m_maxValue;
		// update the data arraylist
		m_data.add(data);
	}

	/** Returns the weight of the data value according to the distribution.
	 * Ranges from 0 to 1.0; all values within the step are given a value of 1.0.
	 * @param data The data value.
	 * @return The weight of the data value.
	 */
	public double getProbability(double data) {
		if (data >= m_minValue && data <= m_maxValue) {
			return 1;
		}
		return 0;
	}
	
	/** Returns a string representation of the SingleStepEstimator
	 * @return A representation of the SingleStepEstimator.
	 */
	public String toString() {
		if (m_minValue > m_maxValue) {
			return "[undefined]";
		}
		DecimalFormat df = new DecimalFormat("0.0000");
		String s = "["+df.format(m_minValue)+","+df.format(m_maxValue)+"]";
		Collections.sort(m_data);
		s += " "+ m_data.toString();
		return s;
	}

}
