/* BlocksGenerator.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package ddpref.data;

import java.io.*;
import java.util.Random;

import weka.core.*;

/*
 * Created on Oct 24, 2005
 * @author Eric Eaton
 */

//TODO Make BlocksGenerator extend weka.datagenerators.Generator
public class BlocksGenerator {
	
	Random randGenerator = null;
	long randomSeed = 0;
	
	public BlocksGenerator() {
		randomSeed = System.currentTimeMillis();
		randGenerator = new Random(randomSeed);
	}

	public BlocksGenerator(long randomSeed) {
		this.randomSeed = randomSeed;
		randGenerator = new Random(randomSeed);
	}
	
	public Instances generateInstances(int numInstances) {
		// initialize the instances header
		Attribute blockIDAtt = new Attribute("blockID");
		Attribute sizeAtt = new Attribute("size");
		/*FastVector colors = new FastVector(7);
		colors.addElement("red");
		colors.addElement("orange");
		colors.addElement("yellow");
		colors.addElement("green");
		colors.addElement("blue");
		colors.addElement("indigo");
		colors.addElement("violet");
		Attribute colorAtt = new Attribute("color", colors);*/
		Attribute colorAtt = new Attribute("color");
		Attribute numSidesAtt = new Attribute("num-sides");
		Attribute binAtt = new Attribute("bin");
		
		FastVector attInfo = new FastVector(5);
		attInfo.addElement(blockIDAtt);
		attInfo.addElement(sizeAtt);
		attInfo.addElement(colorAtt);
		attInfo.addElement(numSidesAtt);
		attInfo.addElement(binAtt);
		
		// initialize the instances set
		Instances data = new Instances("Blocks-"+numInstances+"-"+randomSeed,
				attInfo,numInstances);
		
		// generate the instances
		for (int i=0; i<numInstances; i++) {
			// generate a random size, 1.0 - 100.0
			double size = 1 + randGenerator.nextDouble() * 99;
			// generate a random color
			/*String color = null; 
			switch(randGenerator.nextInt(7)) {
				case 0: color = "red"; break;
				case 1: color = "orange"; break;
				case 2: color = "yellow"; break;
				case 3: color = "green"; break;
				case 4: color = "blue"; break;
				case 5: color = "indigo"; break;
				case 6: color = "violet"; break;
				default: color = "unknown";
			}*/
			int color = randGenerator.nextInt(7);
			// generate a random number of sides, 3 - 20
			int numSides = 3 + randGenerator.nextInt(18);
			// generate a random bin, 0 - 100
			int bin = randGenerator.nextInt(101);
			
			// create the instance
			Instance tempInstance = new Instance(5);
			tempInstance.setValue(blockIDAtt, i);
			tempInstance.setValue(sizeAtt, size);
			tempInstance.setValue(colorAtt, color);
			tempInstance.setValue(numSidesAtt, numSides);
			tempInstance.setValue(binAtt, bin);
			
			// add the instance
			data.add(tempInstance);
		}
		
		return data;
	}
	


	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		// error check the arguments
		if (args.length != 3) {
			System.out.println("Usage: BlocksGenerator <num instances> <random seed> <outputfilename>");
			System.out.println("  <num instances>   The number of instances to generate.");
			System.out.println("  <random seed>     The random seed for generating the data set.");
			System.out.println("  <outputfilename>  The filename to output the instances too.");
			System.exit(1);
		}
		
		// read in the arguments
		int numInstances = 0;
		long randomSeed = 0;
		String outfilename = null;
		File outfile = null; 
		PrintWriter outwriter = null;
		try {
			numInstances = Integer.parseInt(args[0]);
			if (numInstances < 0) throw new Exception();
		} catch (Exception e) {
			System.err.println("<num instances> must be a positive integer.");
			System.exit(1);
		}
		try {
			randomSeed = Integer.parseInt(args[1]);
		} catch (Exception e) {
			System.err.println("<random seed> must be an integer.");
			System.exit(1);
		}
		try {
			outfilename = args[2];
			outfile = new File (outfilename);
			if (outfile.exists()) {
				System.err.println("File \""+outfilename+"\" already exists.");
				System.exit(1);
			}
			outwriter = new PrintWriter(new FileWriter(outfile));
		} catch (Exception e) {
			System.err.println("Error opening file \""+outfilename+"\"for writing.");
			System.exit(1);
		}
		
		// generate the blocks data set
		BlocksGenerator blocksGen = new BlocksGenerator(randomSeed);
		Instances outdata = blocksGen.generateInstances(numInstances);
		
		// save the data to a file, if requested
		if (outwriter != null) {
			try {
				outwriter.println(outdata);
				outwriter.close();
			} catch(Exception e) {
				System.err.println("Error writing data to file: " + e.getMessage());
			}
		}
		
	}

}
