/* LearnPrefs.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/**
 * @author Eric Eaton
 */

package ddpref.experiments;

import java.io.File;
import java.io.FileReader;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;

import weka.core.Instances;
import weka.core.Utils;
import ddpref.WekaUtils;
import ddpref.DDPref;
import ddpref.DDPrefLearner;
import ddpref.Preferences;
import ddpref.experiments.ExperimentUtils;

public class LearnPrefs {	
	
	
	private static String printUsage() {
		String s = "Usage: LearnPrefs <options>\n";
		s += "-D <dataset filename> (required)\n";
		s += "-f <number of folds from the data set (required)\n";
		s += "-P <preferences filename> (required if not using -T)\n";
		s += "-T <subsets filename> (required if not using -P; automatically sets -V and -f)\n";
		s += "-K use top-K (depth-only) instead of depth and diversity\n";
		s += "-k <list of values for the size of the subset> (e.g. -k 5,8,10) (required)\n";
		s += "-a <list of values for alpha> (e.g. -a 0.5,1) (default: 1)\n";
		s += "-t <number of trials> (default: 1)\n";
		s += "-r <random seed>\n";
		s += "-V assume perfect obj fn values of training sets.";
		s += "-d use KDE estimation for learning diversities.";
		return s;
	}
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {

		// command line argument variables
		Instances data = null;
		int numFolds = -1;
		double[] alphaList = null;
		int[] kList = null;
		int numTrials = 1;
		Preferences prefs = null;
		long randomSeed = System.currentTimeMillis();
		Instances[] trainingSubsets = null;
		double[] trainingSubsetsValues = null;
		
		Instances[] dataFolds = null;
		
		int numRandomSubsets = 1000;  // number of random subsets to select from all the data
		
		//int maxNumTrainingSets = -1;
		//boolean pickRandomTrainingSets = false;
		boolean useGivenTrainingSubsets = false;
		boolean useGivenFolds = false;
		boolean useTopK = false;
		boolean usePerfectObjFnValues = false;
		boolean useKDEDiversityEstimation = false;

		boolean verbose = true;
		
		String commandLine = "";
		for (String arg : args)
			commandLine += arg+" ";

		// parse the command line arguments
		try {
			String optionString;

			optionString = Utils.getOption('f', args);
			if (optionString.length() != 0) {
				numFolds = Integer.parseInt(optionString);
			}

			kList = ExperimentUtils.parseIntegerArray(Utils.getOption('k', args));

			optionString = Utils.getOption('a', args);
			if (optionString.length() != 0) {
				alphaList = ExperimentUtils.parseDoubleArray(optionString);
			}

			optionString = Utils.getOption('t', args);
			if (optionString.length() != 0) {
				numTrials = Integer.parseInt(optionString);
			}

			optionString = Utils.getOption('r', args);
			if (optionString.length() != 0) {
				randomSeed = Integer.parseInt(optionString);
			}
			
			//optionString = Utils.getOption('s',args);
			//if (optionString.length() != 0) {
			//	maxNumTrainingSets = Integer.parseInt(optionString);
			//	pickRandomTrainingSets = true;
			//}
			
			if (Utils.getFlag('K',args))
			  useTopK = true;
			
			if (Utils.getFlag('V',args))
				usePerfectObjFnValues = true;
			
			if (Utils.getFlag('d',args))
				useKDEDiversityEstimation = true;

			optionString = Utils.getOption('D', args);
			if (optionString.length() != 0) {
				data = new Instances(new FileReader(new File(optionString)));
			}

			optionString = Utils.getOption('P', args);
			if (optionString.length() != 0) {
				prefs = new Preferences(optionString, new File(optionString));
			}
			
			optionString = Utils.getOption('F', args);
			if (optionString.length() != 0) {
				RankedSubsets rankedSubsets = ExperimentUtils.initializeSubsetsFromFile(data, new File(optionString));
				dataFolds = rankedSubsets.subsets;
				useGivenFolds = true;
				numFolds = rankedSubsets.subsets.length;
				System.out.println("-F option sets numFolds to "+numFolds+".");
			}
			
			optionString = Utils.getOption('T', args);
			if (optionString.length() != 0) {
				RankedSubsets rankedSubsets = ExperimentUtils.initializeSubsetsFromFile(data, new File(optionString));
				trainingSubsets = rankedSubsets.subsets;
				useGivenTrainingSubsets = true;
				//trainingSubsetsValues = rankedSubsets.subsetValues;
				// must use perfectObjFnValues
				System.out.println("-T option sets -V flag automatically.");
				usePerfectObjFnValues = true;
				// numFolds must be same as number of training subsets
				numFolds = trainingSubsets.length;
				System.out.println("-T option sets numFolds to "+numFolds+".");
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			System.err.println("Error from this area of code");
			System.err.println(printUsage());
			System.exit(1);
		}

		// error check command line arguments
		// TODO Check these!!
		try {
			if (kList == null) throw new Exception("k unspecified");
			for (double k : kList)
				if (!ExperimentUtils.valueInRange(k,1,Integer.MAX_VALUE)) throw new Exception("k value out of range");
			for (double a : alphaList)
				if (!ExperimentUtils.valueInRange(a,0,1)) throw new Exception("alpha out of range");
			if (!ExperimentUtils.valueInRange(numFolds,1,Integer.MAX_VALUE)) throw new Exception("numFolds out of range");
			
			if (data == null) throw new Exception("-D must be specified");
			if (prefs == null && trainingSubsets  == null) throw new Exception ("Must specify either -P or -T.");
			if (prefs != null && trainingSubsets  != null) throw new Exception ("Both -T and -P specified.");
			if (numFolds < 0) throw new Exception("-f must be specified");
			//if (maxNumTrainingSets == -1) maxNumTrainingSets = numFolds-1;
		} catch(Exception e) {
			System.err.println("Required value unspecified or out of range: "+e.getMessage());
			System.err.println(printUsage());
			System.exit(1);
		}
		
		// echo command line arguments
		System.out.println("Command line:  "+commandLine);


		try {
			
			// delete the id field now that the training subsets and given folds have been initialized
			data.deleteAttributeAt(0); // delete the instance id from the data
			if (useGivenFolds) {
				for (Instances insts : dataFolds) {
					insts.deleteAttributeAt(0);
				}
			}
			if (useGivenTrainingSubsets) {
				for (Instances insts : trainingSubsets) {
					insts.deleteAttributeAt(0);
				}
			}

			System.out.println();

			Random randGenerator = new Random(randomSeed); // reset the random generator
			
			boolean usingPredefinedPrefs = (prefs!=null);

			// Run the experiment over all values of alpha, and k.
			for (double alpha : alphaList) {
				for (int k : kList) {
					
					// stat collectors
					SummaryStatistics statsP = new SummaryStatisticsImpl();
					SummaryStatistics statsR = new SummaryStatisticsImpl();
					SummaryStatistics[] statsQ = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] correlationPQRand = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] correlationPQBest = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsOverlapp = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsRankP = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsRankQ = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsRelativeP = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsRelativeQ = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsRelativeR = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsEuclideanDistPQ = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsEuclideanDistQR = new SummaryStatisticsImpl[numFolds-1];
					SummaryStatistics[] statsEuclideanDistPR = new SummaryStatisticsImpl[numFolds-1];
					for (int numTrainingSets=0; numTrainingSets<numFolds-1; numTrainingSets++) {
						statsQ[numTrainingSets] = new SummaryStatisticsImpl();
						correlationPQRand[numTrainingSets] = new SummaryStatisticsImpl();
						correlationPQBest[numTrainingSets] = new SummaryStatisticsImpl();
						statsOverlapp[numTrainingSets] = new SummaryStatisticsImpl();
						statsRankP[numTrainingSets] = new SummaryStatisticsImpl();
						statsRankQ[numTrainingSets] = new SummaryStatisticsImpl();
						statsRelativeP[numTrainingSets] = new SummaryStatisticsImpl();
						statsRelativeQ[numTrainingSets] = new SummaryStatisticsImpl();
						statsRelativeR[numTrainingSets] = new SummaryStatisticsImpl();
						statsEuclideanDistPQ[numTrainingSets] = new SummaryStatisticsImpl();
						statsEuclideanDistQR[numTrainingSets] = new SummaryStatisticsImpl();
						statsEuclideanDistPR[numTrainingSets] = new SummaryStatisticsImpl();
					}

					// output the initial prefs
					if (prefs != null) {
						System.out.println("Target Preferences:");
						System.out.println(prefs);
					}


					for (int trial = 0; trial < numTrials; trial++) {

						// compute the number of data items per group and break up the data set
						System.out.println("Running Trial " + trial + " of " + numTrials);
						data.randomize(randGenerator);
							
						if (!useGivenFolds) {
						// divide the data into folds, so that each fold contains one training set
						if (!useGivenTrainingSubsets) {
							// we can just do regular folding of the data if we're picking the training sets later
							dataFolds = ExperimentUtils.separateIntoFolds(data, numFolds);
						} else {
							// get the set of all training instances
							Instances trainingInstances = new Instances(data,0);
							for (Instances instances : trainingSubsets) {
								trainingInstances = WekaUtils.mergeInstances(trainingInstances,instances);
							}
							// get the set of all instances that aren't training instances
							Instances otherData = WekaUtils.getNonMembers(data,trainingInstances);
							// make each fold contain one training set
							dataFolds = ExperimentUtils.separateIntoFolds(otherData, numFolds);
							int i=0;
							for (Instances instances : trainingSubsets) {
								dataFolds[i] = WekaUtils.mergeInstances(dataFolds[i],instances);
								i++;
							}
						}
						} else {
							for (Instances inst : dataFolds) {
								inst.randomize(randGenerator);
							}
						}
						
						DDPref ddpref = new DDPref();
							
						// initialize the training subsets if not specified
						if (!useGivenTrainingSubsets) {
							// generate the subsets from each fold
							trainingSubsets = new Instances[numFolds];
							for (int fold = 0; fold < numFolds; fold++) {
								//generate the top subsets for each fold
								ddpref.setData(dataFolds[fold]);
								trainingSubsets[fold] = ddpref.wrapperSelectSubset(k, prefs);
							}
						}
						
						// gather random testing sets
						Instances[] randomSubsets = new Instances[numFolds];
						for (int fold=0; fold<numFolds; fold++) {
							dataFolds[fold].randomize(randGenerator);
							randomSubsets[fold] = WekaUtils.getSubset(dataFolds[fold],0,k-1);
						}
						Instances[] allDataRandomSubsets = new Instances[numRandomSubsets];
						for (int i=0; i<numRandomSubsets; i++) {
							data.randomize(randGenerator);
							allDataRandomSubsets[i] = WekaUtils.getSubset(data,0,k-1);
						}
						
						// compute P and R if there are target preferences
						if (usingPredefinedPrefs) {
							// compute P
							for (Instances subset : trainingSubsets) {
								statsP.addValue(DDPref.computeObjFunction(subset, prefs));
							}
							// compute R
							for (Instances subset : randomSubsets) {
								statsR.addValue(DDPref.computeObjFunction(subset, prefs));
							}
						}
						

						// perform cross-validation
						for (int heldOutFold = 0; heldOutFold < numFolds; heldOutFold++) {

							// generate the grouping of the training subsets
							ArrayList<Instances> foldTrainingSets = ExperimentUtils.getFoldTrainingSet(trainingSubsets, heldOutFold);
							double[] tempTrainingSetsValues = new double[numFolds - 1];
							int i = 0;
							for (Instances tempTrainingSet : foldTrainingSets) {
								if (usePerfectObjFnValues) {
									tempTrainingSetsValues[i++] = 1.0;
								} else {
									tempTrainingSetsValues[i++] = DDPref.computeObjFunction(tempTrainingSet, prefs);
								}
							}
							// loop through various numbers of training sets
							for (int numTrainingSets=1; numTrainingSets<=foldTrainingSets.size(); numTrainingSets++) {
								
								// get this subset of training sets
								List<Instances> currentTrainingSets = foldTrainingSets.subList(0,numTrainingSets);
								double[] currentTrainingSetsValues = new double[numTrainingSets];
								System.arraycopy(tempTrainingSetsValues,0,currentTrainingSetsValues,0,numTrainingSets);

								// learn preferences from the training sets
								DDPrefLearner ddpreflearner = new DDPrefLearner();
								ddpreflearner.setLearningAlpha(false);
								ddpreflearner.setUseKDEDiversityLearning(useKDEDiversityEstimation);
								ddpreflearner.setAlpha(alpha);
								ddpreflearner.train(currentTrainingSets, currentTrainingSetsValues);
								Preferences currentLearnedPrefs = ddpreflearner.getPreferences();
								
								// output the preferences
								if (verbose) {
									System.out.println("Learned Preferences for Q_"+numTrainingSets+" on fold "+heldOutFold+":");
									System.out.println(currentLearnedPrefs);
									System.out.println();
								}
								
								// compute the best subset from the test data using the learned preferences
								ddpref.setData(dataFolds[heldOutFold]);
								Instances currentSelectedSubset;
								if (useTopK)
								  currentSelectedSubset = ddpref.topSelectSubset(k,currentLearnedPrefs);
								else
								  currentSelectedSubset = ddpref.wrapperSelectSubset(k,currentLearnedPrefs);
								
								if (false) {
									System.out.println("Training subsets: ");
									System.out.print("[ ");
									Instances headerInstances = currentTrainingSets.get(0);
									for (int tempAttIdx = 0; tempAttIdx < headerInstances.numAttributes(); tempAttIdx++) {
										System.out.print(headerInstances.attribute(tempAttIdx).name());
										if (tempAttIdx != (currentSelectedSubset.numAttributes()-1)) 
											System.out.print(",");
									}
									System.out.println(" ]");
									for (Instances insts : currentTrainingSets) {
										System.out.print("{ ");
										for (int tempInstIdx=0; tempInstIdx<insts.numInstances(); tempInstIdx++) {
											System.out.print(insts.instance(tempInstIdx).toString());
											if (tempInstIdx != (insts.numInstances()-1)) 
												System.out.print("; ");
										}
										System.out.println(" }");
									}
									if (useTopK)
									  System.out.println("Top-K Selected subset: ");
									else
									  System.out.println("DD-Select Selected subset: ");
									System.out.print("{ ");
									for (int tempInstIdx=0; tempInstIdx<currentSelectedSubset.numInstances(); tempInstIdx++) {
										System.out.print(currentSelectedSubset.instance(tempInstIdx).toString());
										if (tempInstIdx != (currentSelectedSubset.numInstances()-1)) 
											System.out.print(";\n");
									}
									System.out.println(" }");
												   
									if (verbose) {
										System.out.println();
										System.out.println();
									}
								}
								
								
								// update the stats that use predefined preferences
								if (usingPredefinedPrefs) {
									statsQ[numTrainingSets-1].addValue(DDPref.computeObjFunction(currentSelectedSubset, prefs));

									// update the stats for the correlation
									correlationPQRand[numTrainingSets-1].addValue(Statistics.correlation(allDataRandomSubsets,prefs,currentLearnedPrefs));
									correlationPQBest[numTrainingSets-1].addValue(Statistics.correlation(trainingSubsets,prefs,currentLearnedPrefs));
								}
								
								
								// update the other stats
								statsOverlapp[numTrainingSets-1].addValue(WekaUtils.fmeasure(currentSelectedSubset,trainingSubsets[heldOutFold]));
								statsRelativeP[numTrainingSets-1].addValue(DDPref.computeObjFunction(trainingSubsets[heldOutFold],currentLearnedPrefs));
								statsRelativeQ[numTrainingSets-1].addValue(DDPref.computeObjFunction(currentSelectedSubset,currentLearnedPrefs));
								for (Instances randomSubset : randomSubsets) {
									statsRelativeR[numTrainingSets-1].addValue(DDPref.computeObjFunction(randomSubset,currentLearnedPrefs));
								}
								
								double[] randObjFnValues = new double[allDataRandomSubsets.length];
								int randIdx = 0;
								for (Instances randomSubset : allDataRandomSubsets) {
									randObjFnValues[randIdx++] = DDPref.computeObjFunction(randomSubset,currentLearnedPrefs);
								}
								statsRankP[numTrainingSets-1].addValue(ExperimentUtils.rank(DDPref.computeObjFunction(trainingSubsets[heldOutFold],currentLearnedPrefs),randObjFnValues));
								statsRankQ[numTrainingSets-1].addValue(ExperimentUtils.rank(DDPref.computeObjFunction(currentSelectedSubset,currentLearnedPrefs),randObjFnValues));
								
								statsEuclideanDistPQ[numTrainingSets-1].addValue(WekaUtils.euclideanDistance(trainingSubsets[heldOutFold],currentSelectedSubset));
								statsEuclideanDistQR[numTrainingSets-1].addValue(WekaUtils.euclideanDistance(randomSubsets[heldOutFold],currentSelectedSubset));
								statsEuclideanDistPR[numTrainingSets-1].addValue(WekaUtils.euclideanDistance(trainingSubsets[heldOutFold],randomSubsets[heldOutFold]));

							}
						}
						System.out.println("Done");

					}
					
					System.out.println();
					System.out.println();
						
					// echo command line arguments
					System.out.println("Command line:  "+commandLine);

					// report the statistics
					DecimalFormat df = new DecimalFormat("0.0000");
					String prefsName = (prefs!=null) ? prefs.name() : "UNDEFINED";
					System.out.println("For Alpha = "+alpha+" k = "+k+" prefs = "+prefsName+" over "+numTrials+" trials:");
					System.out.println("\tVp(P)\t\tVp(Q)\t\tVp(R)\t\tr_rand\t\tr_best\t\tOverlapp_PQ\t\tRank_P\t\tRank_Q\t\tVq(P)\t\tVq(Q)\t\tVq(R)\t\tdist(P,Q)\t\tdist(Q,R)\t\tdist(P,R)");
					for (int numTrainingSets=1; numTrainingSets<numFolds; numTrainingSets++) {
						System.out.print(numTrainingSets+"\t");
						
						System.out.print(df.format(statsP.getMean())+"\t"+df.format(statsP.getStandardDeviation())+"\t");
						System.out.print(df.format(statsQ[numTrainingSets-1].getMean())+"\t"+df.format(statsQ[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(statsR.getMean())+"\t"+df.format(statsR.getStandardDeviation())+"\t");
						
						System.out.print(df.format(correlationPQRand[numTrainingSets-1].getMean())+"\t"+df.format(correlationPQRand[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(correlationPQBest[numTrainingSets-1].getMean())+"\t"+df.format(correlationPQBest[numTrainingSets-1].getStandardDeviation())+"\t");
						
						System.out.print(df.format(statsOverlapp[numTrainingSets-1].getMean())+"\t"+df.format(statsOverlapp[numTrainingSets-1].getStandardDeviation())+"\t");
						
						System.out.print(df.format(statsRankP[numTrainingSets-1].getMean())+"\t"+df.format(statsRankP[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(statsRankQ[numTrainingSets-1].getMean())+"\t"+df.format(statsRankQ[numTrainingSets-1].getStandardDeviation())+"\t");
						
						
						System.out.print(df.format(statsRelativeP[numTrainingSets-1].getMean())+"\t"+df.format(statsRelativeP[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(statsRelativeQ[numTrainingSets-1].getMean())+"\t"+df.format(statsRelativeQ[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(statsRelativeR[numTrainingSets-1].getMean())+"\t"+df.format(statsRelativeR[numTrainingSets-1].getStandardDeviation())+"\t");
						
						System.out.print(df.format(statsEuclideanDistPQ[numTrainingSets-1].getMean())+"\t"+df.format(statsEuclideanDistPQ[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.print(df.format(statsEuclideanDistQR[numTrainingSets-1].getMean())+"\t"+df.format(statsEuclideanDistQR[numTrainingSets-1].getStandardDeviation())+"\t");
						System.out.println(df.format(statsEuclideanDistPR[numTrainingSets-1].getMean())+"\t"+df.format(statsEuclideanDistPR[numTrainingSets-1].getStandardDeviation()));
					}
					System.out.println();
				}
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}

	}
	
	
	
	
	
	

}
