/* Statistics.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/**
 * @author Eric Eaton
 */

package ddpref.experiments;

import java.util.Collection;

import ddpref.DDPref;
import ddpref.Preferences;
import weka.core.Instances;

public class Statistics {
	
	/** Computes the correlation between two preferences over a set of instances
	 * Takes a group of subsets and evaluates them with
     * respect to preferences P and to Q.  The objective function values define a
     * profile.  Computes the correlation between P and Q as
     * r = (|S| * \sum_i(P(s_i)*Q(s_i)) - \sum_i(P(s_i)) * \sum_i(Q(s_i)))/
     *      sqrt((N * \sum_i((P(s_i))^2) - (\sum_i(P(s_i)))^2) *
     *           (N * \sum_i((Q(s_i))^2) - (\sum_i(Q(s_i)))^2))
     * This correlation can be tested for statistical significance via t-tests. 
     * See e.g.: http://www.socialresearchmethods.net/kb/statcorr.htm
	 * @return The correlation between the two preferences over the set of subsets
	 */
	public static double correlation(Collection<Instances> subsets, 
			Preferences prefsP, Preferences prefsQ) {
		return correlation((Instances[]) subsets.toArray(new Instances[0]), prefsP, prefsQ);
	}
	
	/** Computes the correlation between two preferences over a set of instances
	 * Takes a group of subsets and evaluates them with
     * respect to preferences P and to Q.  The objective function values define a
     * profile.  Computes the correlation between P and Q as
     * r = (|S| * \sum_i(P(s_i)*Q(s_i)) - \sum_i(P(s_i)) * \sum_i(Q(s_i)))/
     *      sqrt((N * \sum_i((P(s_i))^2) - (\sum_i(P(s_i)))^2) *
     *           (N * \sum_i((Q(s_i))^2) - (\sum_i(Q(s_i)))^2))
     * This correlation can be tested for statistical significance via t-tests. 
     * See e.g.: http://www.socialresearchmethods.net/kb/statcorr.htm
	 * @return The correlation between the two preferences over the set of subsets
	 */
	public static double correlation(Instances[] subsets, 
			Preferences prefsP, Preferences prefsQ) {
		
		// compute each P(s_i) and Q(s_i) and the sums
		double[] pObjFns = new double[subsets.length];
		double[] qObjFns = new double[subsets.length];
		double sumP = 0;
		double sumQ = 0;
		double sumPSquared = 0;
		double sumQSquared = 0;
		double sumPQ = 0;
		int i=0;
		for (Instances subset : subsets) {
			pObjFns[i] = DDPref.computeObjFunction(subset, prefsP);
			qObjFns[i] = DDPref.computeObjFunction(subset, prefsQ);
			sumP += pObjFns[i];
			sumQ += qObjFns[i];
			sumPSquared += pObjFns[i] * pObjFns[i];
			sumQSquared += qObjFns[i] * qObjFns[i];
			sumPQ += pObjFns[i] * qObjFns[i];
		}
		
		// compute the correlation
		double r = (subsets.length * sumPQ - sumP * sumQ) /
	     			Math.sqrt((subsets.length * sumPSquared - sumP*sumP) *
	     					  (subsets.length * sumQSquared - sumQ*sumQ));
		
		return r;
	}

}
